#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>

#include <nllibc.h>

static FILE __nlstdin;
static FILE __nlstdout;
static FILE __nlstderr;

FILE *stdin  = &__nlstdin;
FILE *stdout = &__nlstdout;
FILE *stderr = &__nlstderr;

static char *utoa(unsigned long val)
{
  static char s[32];
  char *p;

  p = s + sizeof(s);
  *(--p) = '\0';
  while (val) {
    *(--p) = "0123456789"[val % 10];
    val /= 10;
  }
  if (*p == '\0')
    *(--p) = '0';

  return p;
}

static char *itoa(long val)
{
  char *p;

  p = utoa((val < 0) ? -val : val);
  if (val < 0)
    *(--p) = '-';

  return p;
}

static char *otoa(unsigned long val)
{
  static char s[32];
  char *p;

  p = s + sizeof(s);
  *(--p) = '\0';
  while (val) {
    *(--p) = "01234567"[val & 0x7];
    val >>= 3;
  }
  if (*p == '\0')
    *(--p) = '0';

  return p;
}

static char *xtoa(unsigned long val, int upper)
{
  static char s[32];
  char *p;
  char *x;

  x = upper ? "0123456789ABCDEF" : "0123456789abcdef";

  p = s + sizeof(s);
  *(--p) = '\0';
  while (val) {
    *(--p) = x[val & 0xf];
    val >>= 4;
  }
  if (*p == '\0')
    *(--p) = '0';

  return p;
}

static int va_sprintf(char *str, const char *format, va_list ap)
{
  char *p = str, *sv;
  const char *f;
  enum {
    MODE_NORMAL = 0,
    MODE_INT = 1,
    MODE_LONG = 2,
    MODE_QUAD = 3,
  } mode;

  mode = MODE_NORMAL;

  for  (f = format; *f; f++) {
    if (mode != MODE_NORMAL) {
      switch (*f) {
      case 'c':
	*(p++) = va_arg(ap, int);
	break;
      case 's':
	sv = va_arg(ap, char *);
	strcpy(p, sv);
	p += strlen(sv);
	break;
      case 'l':
	switch (mode) {
	case MODE_INT:  mode = MODE_LONG; break;
	case MODE_LONG: mode = MODE_QUAD; break;
	default: break;
	}
	break;
      case 'q':
	switch (mode) {
	case MODE_INT:
	case MODE_LONG: mode = MODE_QUAD; break;
	default: break;
	}
	break;
      case 'd':
	switch (mode) {
	case MODE_LONG:
	case MODE_QUAD: sv = itoa(va_arg(ap, long)); break;
	case MODE_INT:
	default:        sv = itoa(va_arg(ap, int)); break;
	}
	strcpy(p, sv);
	p += strlen(sv);
	break;
      case 'u':
	switch (mode) {
	case MODE_LONG:
	case MODE_QUAD: sv = utoa(va_arg(ap, unsigned long));
	case MODE_INT:
	default:        sv = utoa(va_arg(ap, unsigned int));
	}
	strcpy(p, sv);
	p += strlen(sv);
	break;
      case 'o':
	switch (mode) {
	case MODE_LONG:
	case MODE_QUAD: sv = otoa(va_arg(ap, unsigned long));
	case MODE_INT:
	default:        sv = otoa(va_arg(ap, unsigned int));
	}
	strcpy(p, sv);
	p += strlen(sv);
	break;
      case 'x':
	switch (mode) {
	case MODE_LONG:
	case MODE_QUAD: sv = xtoa(va_arg(ap, unsigned long), 0);
	case MODE_INT:
	default:        sv = xtoa(va_arg(ap, unsigned int), 0);
	}
	strcpy(p, sv);
	p += strlen(sv);
	break;
      case 'X':
	switch (mode) {
	case MODE_LONG:
	case MODE_QUAD: sv = xtoa(va_arg(ap, unsigned long), 1);
	case MODE_INT:
	default:        sv = xtoa(va_arg(ap, unsigned int), 1);
	}
	strcpy(p, sv);
	p += strlen(sv);
	break;
      case 'p':
	strcpy(p, "0x");
	p += 2;
	sv = xtoa((unsigned long)va_arg(ap, char *), 0);
	strcpy(p, sv);
	p += strlen(sv);
	break;
      case '%':
	*(p++) = '%';
	break;
      default:
	break;
      }
      mode = MODE_NORMAL;
      continue;
    }

    if (*f == '%') {
      mode = MODE_INT;
      continue;
    }

    *(p++) = *f;
  }

  *p = '\0';

  return 0;
}

int sprintf(char *str, const char *format, ...)
{
  va_list arg;
  int done;

  va_start(arg, format);
  done = va_sprintf(str, format, arg);
  va_end(arg);

  return done;
}

int printf(const char *format, ...)
{
  va_list arg;
  int done;
  char str[1024*16];

  va_start(arg, format);
  done = va_sprintf(str, format, arg);
  va_end(arg);

  write(1, str, strlen(str));

  return done;
}

int puts(const char *str)
{
  write(1, str, strlen(str));
  write(1, "\n", 1);
  return 0;
}

char *fgets(char *str, int size, FILE *stream)
{
  int r;
  r = read(0, str, size);
  if (r <= 0)
    return NULL;
  return str;
}

int fputs(const char *str, FILE *stream)
{
  return write(1, str, strlen(str));
}
